
/*
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 *  
 */

package com.sunlabs.archsim.fleet.ships.stride;

import java.util.HashMap;
import java.util.Map;

import com.sunlabs.archsim.base.common.Parameters;
import com.sunlabs.archsim.base.simulation.Delay;
import com.sunlabs.archsim.gasp.Component;
import com.sunlabs.archsim.gasp.State;
import com.sunlabs.archsim.gasp.Terminal;
import com.sunlabs.archsim.gasp.terminals.InputTerminal;
import com.sunlabs.archsim.gasp.terminals.OutputTerminal;

/**
 * 
 * @author Igor
 *
 * A stride ship.
 * Used as a loop counter.
 * Set-up intterface: 	
 * 	start = input, the starting number (the first counter value produced)
 * 	step - input, the increment
 * 	count - input, the number of counts
 * 	done - output, token confirming that the count has completed
 * Each of the above has an ack.
 * 
 * For each "loop", one set-up takes place.
 * 
 *  Working interface:
 *  	counter - output, the current count
 *  next - input, a token requestring the next counter value
 *  Each of the above has an ack
 */

public class Stride extends Component {
	

	/* input - start value for the count */
	private Terminal startValueTerm;
	/* output - ack for the start value */
	private Terminal ackStartValueTerm;
	
	/* input - step size */
	private Terminal stepTerm;
	/* output - ack step size */
	private Terminal ackStepTerm;
	
	/* input - number of steps */
	private Terminal stepCountTerm;
	/* output - ack for stepCount */
	private Terminal ackStepCountTerm;
	
	/* input - token to produce the next counter */
	private Terminal nextTerm= null;
	/* output - LOCAL ack that the token has been received */
	private Terminal ackNextTerm= null;
	
	/* output - the counter */
	private Terminal counterTerm= null;
	/** input - LOCAL ack that the counter has been received */
	private Terminal ackCounterTerm;
	
	/* output - cycle is done, ack is an input */
	private Terminal doneTerm;
	private Terminal ackDoneTerm;
	
	/* input - internal done signal, ack is an output */
	private Terminal intDoneTermIn;
	private Terminal ackIntDoneTermIn;

	/* output - internal done signal, ack is an input */
	private Terminal intDoneTermOut;
	private Terminal ackIntDoneTermOut;
	
	/* states that connect setup and done actions */
	private State intDoneState;
	private State intAckDoneState;
	
	/* this mape will be used in getTerminal */
	Map<String, Terminal> terminalNameMap;
	
	// three actions that comprise the stride
	DoneAction doneAction;
	CountingAction countingAction;
	SetupAction setupAction;
	
	public Stride(String name, Parameters params) {
		super(name, null);
		
		// process the parameters
		Delay countDelayParam= null;
		Delay ackDelayParam= null;
		if (params != null) {
			for (int i= 0; i<params.size(); i++) {
				String pName= params.getName(i);
				String sValue= params.getValue(i);
		
				if (pName.equals(StrideDefaults.COUNTER_DELAY) 
						|| pName.equals(StrideDefaults.STRIDE_COUNTER_DELAY)) {
					try {
						int d= Integer.parseInt(sValue);
						countDelayParam= new Delay(d);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for "
								+ StrideDefaults.STRIDE_COUNTER_DELAY
								+ ": " + sValue
								+ ". Value is IGNORED.");						
					}					
					
				}
				else if (pName.equals(StrideDefaults.ACK_DELAY) 
						|| pName.equals(StrideDefaults.STRIDE_ACK_DELAY)) {
					try {
						int d= Integer.parseInt(sValue);
						countDelayParam= new Delay(d);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for "
								+ StrideDefaults.STRIDE_ACK_DELAY
								+ ": " + sValue
								+ ". Value is IGNORED.");						
					}					
					
				}
				else {
					director.error("Component: " + name + ", id= " + getID() 
							+ ": Attribute " + pName + " with value " + sValue
							+ " is ignored");
				}				
		
			} // for
		} // if params != null		
		
		buildStride();
		
		// apply parameters is needed
		if (countDelayParam != null) {
			countingAction.count.setDelay(countDelayParam);
		}
		if (ackDelayParam != null) {
			countingAction.ack.setDelay(ackDelayParam);
			setupAction.ack.setDelay(ackDelayParam);
			setupAction.ackAndTrigger.setDelay(ackDelayParam);
			doneAction.doneFunction.setDelay(ackDelayParam);
		    doneAction.doneAndDisableFunction.setDelay(ackDelayParam);
		}
		
	} // Stride constructor

	protected void buildStride() {
		terminalNameMap= new HashMap<String, Terminal>();
		
		// make the actions
		// Note - the creation order matters so that triggers are available when needed
		doneAction= new DoneAction(StrideDefaults.DONE_ACTION);
		addMember(doneAction);
		
		// counting action needs enable trigger from the done action
		countingAction= new CountingAction(StrideDefaults.COUNTING_ACTION,
				doneAction.getEnableTriggerCommand());
		addMember(countingAction);
		
		setupAction= new SetupAction(StrideDefaults.SETUP_ACTION,
				countingAction.getEnableTriggerCommand());
		addMember(setupAction);
		
		
		// terminals for setupAction, prepare the terminalNameMap
		startValueTerm= new InputTerminal(StrideDefaults.START_VALUE, this);
		terminalNameMap.put(StrideDefaults.START_VALUE, startValueTerm);
		
		ackStartValueTerm= new OutputTerminal(StrideDefaults.ACK_START_VALUE, this);
		terminalNameMap.put(StrideDefaults.ACK_START_VALUE, ackStartValueTerm);		
		
		stepTerm= new InputTerminal(StrideDefaults.STEP, this);
		terminalNameMap.put(StrideDefaults.STEP, stepTerm);
		
		ackStepTerm= new OutputTerminal(StrideDefaults.ACK_STEP, this);
		terminalNameMap.put(StrideDefaults.ACK_STEP, ackStepTerm);
		
		stepCountTerm= new InputTerminal(StrideDefaults.STEP_COUNT, this);
		terminalNameMap.put(StrideDefaults.STEP_COUNT, stepCountTerm);
		
		ackStepCountTerm= new OutputTerminal(StrideDefaults.ACK_STEP_COUNT, this);
		terminalNameMap.put(StrideDefaults.ACK_STEP_COUNT, ackStepCountTerm);
		
		intDoneTermIn= new InputTerminal(StrideDefaults.INT_CYCLE_DONE_IN, this);
		terminalNameMap.put(StrideDefaults.INT_CYCLE_DONE_IN, intDoneTermIn);
		
		ackIntDoneTermIn= new OutputTerminal(StrideDefaults.INT_ACK_CYCLE_DONE_IN, this);
		terminalNameMap.put(StrideDefaults.INT_ACK_CYCLE_DONE_IN, ackIntDoneTermIn);
		
		// attach the terminals
		setupAction.attach(startValueTerm);
		setupAction.attach(ackStartValueTerm);
		setupAction.attach(stepTerm);
		setupAction.attach(ackStepTerm);
		setupAction.attach(stepCountTerm);
		setupAction.attach(ackStepCountTerm);
		setupAction.attach(intDoneTermIn);
		setupAction.attach(ackIntDoneTermIn);
		
		
		// terminals for the countingAction, keep filling the map
		nextTerm= new InputTerminal(StrideDefaults.NEXT, this);
		terminalNameMap.put(StrideDefaults.NEXT, nextTerm);
		
		ackNextTerm= new OutputTerminal(StrideDefaults.ACK_NEXT, this);
		terminalNameMap.put(StrideDefaults.ACK_NEXT, ackNextTerm);
		
		counterTerm= new OutputTerminal(StrideDefaults.COUNTER, this);
		terminalNameMap.put(StrideDefaults.COUNTER, counterTerm);
		
		ackCounterTerm= new InputTerminal(StrideDefaults.ACK_COUNTER, this);
		terminalNameMap.put(StrideDefaults.ACK_COUNTER, ackCounterTerm);
		
		// attach the terminals
		countingAction.attach(nextTerm);
		countingAction.attach(ackNextTerm);
		countingAction.attach(counterTerm);
		countingAction.attach(ackCounterTerm);
		
		// terminals for the doneAction
		doneTerm= new OutputTerminal(StrideDefaults.DONE, this);
		terminalNameMap.put(StrideDefaults.DONE, doneTerm);

		ackDoneTerm= new InputTerminal(StrideDefaults.ACK_DONE, this);
		terminalNameMap.put(StrideDefaults.ACK_DONE, ackDoneTerm);
		
		intDoneTermOut= new OutputTerminal(StrideDefaults.INT_CYCLE_DONE_OUT, this);
		terminalNameMap.put(StrideDefaults.INT_CYCLE_DONE_OUT, intDoneTermOut);

		ackIntDoneTermOut= new InputTerminal(StrideDefaults.INT_ACK_CYCLE_DONE_OUT, this);
		terminalNameMap.put(StrideDefaults.INT_ACK_CYCLE_DONE_OUT, ackIntDoneTermOut);
	
		// attach the terminals
		doneAction.attach(doneTerm);
		doneAction.attach(ackDoneTerm);
		doneAction.attach(intDoneTermOut);
		doneAction.attach(ackIntDoneTermOut);
		
		// make the states that connect the setup action and the done action
		intDoneState= new State(StrideDefaults.INT_CYCLE_DONE_STATE, this);
		intAckDoneState= new State(StrideDefaults.INT_ACK_CYCLE_DONE_STATE, this);
		// initialize the ack state!!!
		intDoneState.setInitiallyActive(null);
		
		// attach the terminals to the internal states - one input and one output
		intDoneState.attach(intDoneTermOut);
		intDoneState.attach(intDoneTermIn);
		
		intAckDoneState.attach(ackIntDoneTermIn);
		intAckDoneState.attach(ackIntDoneTermOut);		
	}
	
	/** get the terminal name */
	public Terminal getTerminal(String name) {
		// use the map, this is why we made it
		return terminalNameMap.get(name);
	}

}
